@@ -29,8 +29,8 @@ def filter_op_not(flt):
 
 def filter_match(val, getter):
     if type(val) is str and len(val) >= 2 and val[0] == '/' and val[-1] == '/':
-        r = re.compile(val)
-        return lambda value: bool(r.match(str(getter(value) or '')))
+        r = re.compile(val[1:-1])
+        return lambda value: bool(r.search(str(getter(value) or '')))
     else:
         return lambda value: getter(value) == val
 
@@ -45,6 +45,15 @@ def make_filter(values, flt_fn, getter):
     return filter_op_or([flt_fn(v, getter) for v in ensure_list(values)])
 
 
+def update_type_getter(value):
+    status = value.props.getProperty('event.change.status')
+    if status == 'MERGED':
+        return 'merged'
+    elif status == 'TAGGED':
+        return 'tagged'
+    return 'new'
+
+
 FILTERS = {
     'branch': (filter_match, operator.attrgetter('branch')),
     'branches': (filter_match, operator.attrgetter('branch')),
@@ -52,6 +61,8 @@ FILTERS = {
     'comments': (filter_match, operator.attrgetter('comments')),
     'file': (filter_fnmatch, operator.attrgetter('files')),
     'files': (filter_fnmatch, operator.attrgetter('files')),
+    'status': (filter_match, update_type_getter),
+    'tag': (filter_match, lambda value: value.props.getProperty('event.change.tag')),
 }
 
 
@@ -69,12 +80,9 @@ def make_filters(filters, is_or=False):
                 result.append(filter_op_not(make_filters(values)))
             elif name == 'or':
                 result.append(make_filters(values, True))
-            elif name.startswith('prop_'):
-                name = name[5:]
+            else:
                 getter = lambda value: value.props.getProperty(name)
                 result.append(make_filter(values, filter_match, getter))
-            else:
-                raise Exception(f'Unknown filter: {name}')
 
     if is_or:
         return filter_op_or(list(filter(None, result)))
