@@ -1,4 +1,5 @@
 import os.path
+import itertools
 import json
 import yaml
 import re
@@ -14,13 +15,18 @@ from twisted.python.failure import Failure
 
 from buildbot_pipeline import junit, utils, filters, file_store, build
 
+DEFAULT_STEPSDIR = 'buildbot'
+HIDDEN = 'hidden'
+
 
 def process_interpolate(value):
     if type(value) is str:
         if value.startswith('!'):
             return properties.Interpolate(value[1:])
-        elif value.startswith(r'\!'):
-            return value[2:]
+        elif value.startswith(r'@'):
+            return value[1:]
+        elif '%(' in value:
+            return properties.Interpolate(value)
         else:
             return value
     elif type(value) is list:
@@ -45,13 +51,23 @@ def gen_steps(step, data):
         data['env'] = step.build.pipeline_env.copy()
         data['env'].update(step_env)
         data = process_interpolate(data)
-        data['env']['BUILD_ID'] = properties.Interpolate('%(prop:root_buildnumber:-%(prop:buildnumber)s)s')
+        data['env']['BUILD_ID'] = properties.Interpolate('%(prop:pipeline_buildnumber:-%(prop:buildnumber)s)s')
         data['env']['WORKSPACE'] = properties.Interpolate('%(prop:builddir)s')
         return DynamicStep(**data)
     elif 'steps' in data:
         return [gen_steps(step, it) for it in data['steps']]
     elif 'parallel' in data:
-        return Parallel(data['parallel'])
+        info = data['parallel']
+        if type(info) is list:
+            info = {'steps': info}
+
+        matrix = info.pop('matrix', None)
+        steps = info.pop('steps', [])
+        if matrix:
+            steps.insert(0, {'matrix': matrix})
+
+        info.pop('inner', None)
+        return Parallel(list(matrix_steps(steps)), **info)
     elif 'git' in data:
         data = process_interpolate(data)
         data.pop('git')
@@ -61,6 +77,28 @@ def gen_steps(step, data):
     raise Exception(f'Unknown step {data}')
 
 
+def matrix_steps(steps):
+    for info in steps:
+        if 'matrix' in info:
+            step_desc = info['matrix']
+            name = step_desc.pop('name', None)
+            all_params = step_desc.pop('params', [])
+            for params in utils.ensure_list(all_params):
+                params.pop('workername', None)
+                params = list(params.items())
+                params_keys = [it[0] for it in params]
+                params_values = [utils.ensure_list(it[1]) for it in params]
+                for pvals in itertools.product(*params_values):
+                    props = dict(zip(params_keys, pvals))
+                    s = step_desc.copy()
+                    s['properties'] = s.get('properties', {}).copy()
+                    s['properties'].update(props)
+                    s['name'] = name.format(**props) if name else '-'.join(map(str, pvals))
+                    yield s
+        else:
+            yield info
+
+
 class Parallel(Trigger):
     def __init__(self, steps_info, inner=True, **kwargs):
         kwargs.setdefault('waitForFinish', True)
@@ -69,7 +107,7 @@ class Parallel(Trigger):
         self.correct_names = [it['name'] for it in steps_info]
         self.inner = inner
 
-    # @defer.inlineCallbacks
+    @defer.inlineCallbacks
     def getSchedulersAndProperties(self):
         if self.inner:
             builder_name = self.getProperty('parent_builder_name') or self.getProperty('virtual_builder_name') or self.getProperty('buildername')
@@ -87,9 +125,17 @@ class Parallel(Trigger):
                 'unimportant': False
             }
 
+            passthrough_propnames = self.getProperty('pipeline_passthrough_props') or []
+            props = {k: self.getProperty(k) for k in passthrough_propnames}
+            props.update(it.get('properties', {}))
+            props = yield self.render(process_interpolate(props))
+            s['props_to_set'].update(props)
+
             if self.inner:
                 s['props_to_set'].update(
                     virtual_builder_name=f"{builder_name}/{it['name']}",
+                    virtual_builder_title=it['name'],
+                    virtual_builder_tags=[HIDDEN],
                     workername=worker_name,
                     parent_builder_name=builder_name,
                     pipeline_builddir=builddir,
@@ -99,6 +145,7 @@ class Parallel(Trigger):
             else:
                 s['props_to_set'].update(
                    virtual_builder_name=it['name'],
+                   virtual_builder_tags=[],
                    pipeline_builder_prefix='~prop-builder',
                    reporters='gerrit',
                    pipeline_concurrency=str(it.get('concurrency', 1)),
@@ -155,10 +202,16 @@ class GatherBuilders(buildstep.BuildStep):
     @defer.inlineCallbacks
     def run(self):
         workdir = os.path.join(self.getProperty('builddir'), self.workdir)
+
+        build_props = self.getProperty('pipeline_build_props', {})
+        forced_builders = build_props.get('builders')
+        common_props = build_props.get('common_props', {})
+        builder_props = build_props.get('builder_props', {})
+
         result = results.SUCCESS
         step_info = []
         repo_path = Path(workdir)
-        buildbot_path = repo_path / 'buildbot'
+        buildbot_path = repo_path / self.getProperty('pipeline_stepsdir', DEFAULT_STEPSDIR)
         for it in buildbot_path.glob('**/*.yaml'):
             it_repo_path = str(it.relative_to(repo_path))
             name, _, _ = str(it.relative_to(buildbot_path)).rpartition('.')
@@ -182,10 +235,16 @@ class GatherBuilders(buildstep.BuildStep):
                         start_build = False
                         skip_reason = 'disabled'
 
+                    if start_build is None and forced_builders and name in forced_builders:
+                        start_build = True
+
                     if start_build is None and it_repo_path in changes[0].files:
                         start_build = True
 
                     if start_build is None and 'filter' in step:
+                        if 'status' not in step['filter']:
+                            step['filter']['status'] = 'new'
+
                         try:
                             flt = filters.make_filters(step['filter'])
                         except Exception as e:
@@ -200,6 +259,11 @@ class GatherBuilders(buildstep.BuildStep):
                                 skip_reason = 'filter'
 
                     if start_build:
+                        props = step.get('properties', {}).copy()
+                        props.update(common_props)
+                        props.update(builder_props.get(name, {}))
+                        props['pipeline_passthrough_props'] = list(props) + ['pipeline_passthrough_props']
+                        step['properties'] = props
                         step_info.append(step)
                     else:
                         yield self.addCompleteLog(name, f'skipped by: {skip_reason}')
@@ -260,6 +324,7 @@ class DynamicStep(buildstep.ShellMixin, buildstep.BuildStep):
 
     @defer.inlineCallbacks
     def handleJUnit(self, desc):
+        desc = yield self.render(desc)
         if type(desc) is dict:
             label = desc.get('label')
             src = desc.get('src')
@@ -289,10 +354,11 @@ class DynamicStep(buildstep.ShellMixin, buildstep.BuildStep):
 
     @defer.inlineCallbacks
     def handleUpload(self, desc):
+        desc = yield self.render(desc)
         bname = build.builder_name_to_path(self.getProperty('virtual_builder_name') or self.getProperty('buildername'))
         bnum = self.getProperty('pipeline_buildnumber') or self.getProperty('buildnumber')
 
-        build_storage_path = os.path.realpath(os.path.join(file_store.app.path, bname, str(bnum)))
+        build_storage_path = os.path.realpath(os.path.join(file_store.ep.path, bname, str(bnum)))
         if desc.get('dest'):
             dest = os.path.realpath(os.path.join(build_storage_path, desc['dest']))
             if len(os.path.commonpath([dest, build_storage_path])) < len(build_storage_path):
@@ -303,7 +369,7 @@ class DynamicStep(buildstep.ShellMixin, buildstep.BuildStep):
         cmd = MultipleFileUpload(
             workersrcs=utils.ensure_list(desc['src']),
             glob=True,
-            url=f'file-store/{bname}/{bnum}/' + desc.get('link', ''),
+            url=f'/file-store/{bname}/{bnum}/' + desc.get('link', ''),
             urlText=desc.get('label'),
             masterdest=dest)
 
